{{
  vga_40x18_rom_hc.spin by Roger Williams

  This is the helper cog for a ROM font 43 by 18 character tile
  display.  The raster is drawn by a raster engine which signals
  which line it is working on via the lineptr.  Two instances
  of this cog pre-build lines of pixels, one working on even
  lines and one on odd.

}}
VAR
  long  cog0, cog1

PUB start(screen, pixelbuf, lineptr, userfontptr) 

  _screen     := screen
  _pixbuf     := pixelbuf
  _helpsync   := lineptr
  _user_chars := userfontptr

  cog0 := cognew(@feeder, %0000) + 1
  cog1 := cognew(@feeder, %1100) + 1

  ifnot result := cog0 and cog1
    stop

PUB stop

  if cog0
    cogstop(cog0~ -1)
  if cog1
    cogstop(cog1~ -1)

DAT             org     0

feeder          jmp    {#:rac0f_000_0}#:skip    ' skip table
                jmp     #:rac0f_000_1
                long    -1[2]
                jmp     #:rac0f_001_0
                jmp     #:rac0f_001_1
                long    -1[2]
                jmp     #:rac0f_010_0
                jmp     #:rac0f_010_1
                long    -1[2]
                jmp     #:rac0f_011_0
                jmp     #:rac0f_011_1
                long    -1[2]
                jmp     #:rac0f_100_0
                jmp     #:rac0f_100_1
                long    -1[2]
                jmp     #:rac0f_101_0
                jmp     #:rac0f_101_1
                long    -1[2]
                jmp     #:rac0f_110_0
                jmp     #:rac0f_110_1
                long    -1[2]
                jmp     #:rac0f_111_0
                jmp     #:rac0f_111_1

                jmp     #:rac0f_000_0

:skip           mov     feeder, $-1             ' restore entry 0
                cmp     par, par wz             ' even/odd
        if_z    xor     :cue, opposite          ' nc -> c
        if_z    movs    :out, #0                ' buffer offset


:loop           rdword  thisline, _helpsync     ' new line?
                cmp     thisline, lastline wz
        if_z    jmp     #:loop

                mov     lastline, thisline      ' new line is now current line
                shr     lastline, #1 wc,nr      ' get lsb in c
:cue    if_nc   jmp     #:loop                  ' only process this if it's our cue

                add     thisline, #2            ' process the next line, not the
                cmp     thisline, lines wc      ' one currently being displayed
        if_nc   sub     thisline, lines

                ' initialize pointer to 1st character of line in screen

                mov     screenptr, thisline     ' pixel line 0-575
                shr     screenptr, #5           ' discard tile line 0-31
                shl     screenptr, #2           ' text line * 50 words * 2 bytes/word
                mov     temp, screenptr         ' *4
                shl     screenptr, #3           ' *32
                add     temp, screenptr         ' *36
                shl     screenptr, #1           ' *64
                add     screenptr, temp         ' *100
                add     screenptr, _screen      ' offset to _screen
                                                        
                ' initialize line-within-tile memory pointer                                               
                                                                                                           
                mov     tileline, thisline                                                                 
                and     tileline, #%0_0001_1111 ' %00000000_00000000_00000000_000##### isolate y 0-31
                shl     tileline, #2 + 17       ' long pointer                                             
                                                                                                           
                ' initialize output pointers                                                               
                                                                                                           
:out            mov     pixbufptr, #columns * 4                                                            
                add     pixbufptr, _pixbuf                                                                 
                  '                                                                                        
                mov     x, #columns             ' set horizontal tiles per line                            

                ' tile-pixel conversion loop

:tile           rdword  tile, screenptr         ' read tile
                ror     tile, #8                ' %hCCCCCCC_00000000_00000000_000rac0f
                add     tile, tileline{<<17}    ' %hCCCCCCC_tline000_00000000_000rac0f
                jmp     tile


:rac0f_000_0    ror     tile, #18               ' %00000000_00000rac_0fhCCCCC_CCtline0
                add     tile, _user_chars       ' add user offset

                rdlong  pixels, tile           
                and     pixels, loqbit          ' remove unwanted interleaved char
                add     screenptr, #2           ' point to next tile

                wrlong  pixels, pixbufptr
                add     pixbufptr, #4
                djnz    x, #:tile               ' another tile in line?
                                 
                jmp     #:loop
                        

:rac0f_000_1    ror     tile, #17               ' %00000000_0000rac0_fhCCCCCC_Ctline00

                rdlong  pixels, tile           
                and     pixels, loqbit          ' remove unwanted interleaved char
                add     screenptr, #2           ' point to next tile

                wrlong  pixels, pixbufptr
                add     pixbufptr, #4
                djnz    x, #:tile               ' another tile in line?

                jmp     #:loop


:rac0f_001_0    ror     tile, #18               ' %00000000_00000rac_0fhCCCCC_CCtline0
                add     tile, _user_chars       ' add user offset

                rdlong  pixels, tile           
                shr     pixels, #1              ' based on LSB, get all desired bits %%1
                and     pixels, loqbit          ' remove unwanted interleaved char

                wrlong  pixels, pixbufptr
                add     pixbufptr, #4
                add     screenptr, #2           ' point to next tile

                djnz    x, #:tile               ' another tile in line?

                jmp     #:loop


:rac0f_001_1    ror     tile, #17               ' %00000000_0000rac0_fhCCCCCC_Ctline00
                add     screenptr, #2           ' point to next tile

                rdlong  pixels, tile           
                shr     pixels, #1              ' based on LSB, get all desired bits %%1
                and     pixels, loqbit          ' remove unwanted interleaved char

                wrlong  pixels, pixbufptr
                add     pixbufptr, #4
                djnz    x, #:tile               ' another tile in line?

                jmp     #:loop


:rac0f_010_0    ror     tile, #18               ' %00000000_00000rac_0fhCCCCC_CCtline0
                add     tile, _user_chars       ' add user offset

                rdlong  pixels, tile           
                and     pixels, loqbit          ' remove unwanted interleaved char
                shl     pixels, #1              ' alt color %%2

                wrlong  pixels, pixbufptr
                add     pixbufptr, #4
                add     screenptr, #2           ' point to next tile

                djnz    x, #:tile               ' another tile in line?

                jmp     #:loop


:rac0f_010_1    ror     tile, #17               ' %00000000_0000rac0_fhCCCCCC_Ctline00
                add     screenptr,#2            ' point to next tile

                rdlong  pixels, tile           
                and     pixels, loqbit          ' remove unwanted interleaved char
                shl     pixels, #1              ' alt color %%2

                wrlong  pixels, pixbufptr
                add     pixbufptr, #4
                djnz    x, #:tile               ' another tile in line?

                jmp     #:loop


:rac0f_011_0    ror     tile, #18               ' %00000000_00000rac_0fhCCCCC_CCtline0
                add     tile, _user_chars       ' add user offset

                rdlong  pixels, tile           
                shr     pixels, #1              ' based on LSB, get all desired bits %%1
                and     pixels, loqbit          ' remove unwanted interleaved char

                shl     pixels, #1              ' alt color %%2
                add     screenptr, #2           ' point to next tile

                wrlong  pixels, pixbufptr
                add     pixbufptr, #4
                djnz    x, #:tile               ' another tile in line?

                jmp     #:loop


:rac0f_011_1    ror     tile, #17               ' %00000000_0000rac0_fhCCCCCC_Ctline00          
                                                                                        
                rdlong  pixels, tile                                                            
                shr     pixels, #1              ' based on LSB, get all desired bits %%1        
                and     pixels, loqbit          ' remove unwanted interleaved char              
                                                                                        
                shl     pixels, #1              ' alt color %%2                                 
                add     screenptr, #2           ' point to next tile                            
                                                                                        
                wrlong  pixels, pixbufptr                                                       
                add     pixbufptr, #4                                                           
                djnz    x, #:tile               ' another tile in line?                         
                                                                                        
                jmp     #:loop                                                                  


:rac0f_100_0    ror     tile, #18               ' %00000000_00000rac_0fhCCCCC_CCtline0
                add     tile, _user_chars       ' add user offset                     
                                                                                      
                rdlong  pixels, tile                                                  
                xor     pixels, loqbit                                                
                and     pixels, loqbit          ' remove unwanted interleaved char    
                                                                                      
                wrlong  pixels, pixbufptr                                             
                add     pixbufptr, #4                                                 
                add     screenptr, #2           ' point to next tile                  
                                                                                      
                djnz    x, #:tile               ' another tile in line?               
                                                                                      
                jmp     #:loop                                                        


:rac0f_100_1    ror     tile, #17               ' %00000000_0000rac0_fhCCCCCC_Ctline00
                add     screenptr, #2           ' point to next tile                  
                                                                                      
                rdlong  pixels, tile                                                  
                xor     pixels, loqbit                                                
                and     pixels, loqbit          ' remove unwanted interleaved char    
                                                                                      
                wrlong  pixels, pixbufptr                                             
                add     pixbufptr, #4                                                 
                djnz    x, #:tile               ' another tile in line?               
                                                                                      
                jmp     #:loop                                                        


:rac0f_101_0    ror     tile, #18               ' %00000000_00000rac_0fhCCCCC_CCtline0  
                add     tile, _user_chars       ' add user offset                       
                                                                                        
                rdlong  pixels, tile                                                    
                shr     pixels, #1              ' based on LSB, get all desired bits %%1
                xor     pixels, loqbit                                                  
                                                                                        
                and     pixels, loqbit          ' remove unwanted interleaved char      
                add     screenptr, #2           ' point to next tile                    
                                                                                        
                wrlong  pixels, pixbufptr                                               
                add     pixbufptr, #4                                                   
                djnz    x, #:tile               ' another tile in line?                 
                                                                                        
                jmp     #:loop                                                          


:rac0f_101_1    ror     tile, #17               ' %00000000_0000rac0_fhCCCCCC_Ctline00  
                add     screenptr, #2           ' point to next tile                    
                                                                                        
                rdlong  pixels, tile                                                    
                shr     pixels, #1              ' based on LSB, get all desired bits %%1
                xor     pixels, loqbit                                                  
                                                                                        
                and     pixels, loqbit          ' remove unwanted interleaved char      
                                                                                        
                wrlong  pixels, pixbufptr                                               
                add     pixbufptr, #4                                                   
                djnz    x, #:tile               ' another tile in line?                 
                                                                                        
                jmp     #:loop                                                          


:rac0f_110_0    ror     tile, #18               ' %00000000_00000rac_0fhCCCCC_CCtline0
                add     tile, _user_chars       ' add user offset                     
                                                                                      
                rdlong  pixels, tile                                                  
                xor     pixels, loqbit                                                
                and     pixels, loqbit          ' remove unwanted interleaved char    
                                                                                      
                shl     pixels, #1              ' alt color %%2                       
                add     screenptr, #2           ' point to next tile                  
                                                                                      
                wrlong  pixels, pixbufptr                                             
                add     pixbufptr, #4                                                 
                djnz    x, #:tile               ' another tile in line?               
                                                                                      
                jmp     #:loop                                                        
                                                                                      
                                                                                      
:rac0f_110_1    ror     tile, #17               ' %00000000_0000rac0_fhCCCCCC_Ctline00
                                                                                      
                rdlong  pixels, tile                                                  
                xor     pixels, loqbit                                                
                and     pixels, loqbit          ' remove unwanted interleaved char    
                                                                                      
                shl     pixels, #1              ' alt color %%2                       
                add     screenptr, #2           ' point to next tile                  
                                                                                      
                wrlong  pixels, pixbufptr                                             
                add     pixbufptr, #4                                                 
                djnz    x, #:tile               ' another tile in line?               
                                                                                      
                jmp     #:loop                                                        


:rac0f_111_0    ror     tile, #18               ' %00000000_00000rac_0fhCCCCC_CCtline0  
                add     tile, _user_chars       ' add user offset                       
                                                                                        
                rdlong  pixels, tile                                                    
                shr     pixels, #1              ' based on LSB, get all desired bits %%1
                xor     pixels, loqbit                                                  
                                                                                        
                and     pixels, loqbit          ' remove unwanted interleaved char      
                shl     pixels, #1              ' alt color %%2                         
                add     screenptr, #2           ' point to next tile                    
                                                                                        
                wrlong  pixels, pixbufptr                                               
                add     pixbufptr, #4                                                   
                djnz    x, #:tile               ' another tile in line?                 
                                                                                        
                jmp     #:loop                                                          
                                                                                        
                                                                                        
:rac0f_111_1    ror     tile, #17               ' %00000000_0000rac0_fhCCCCCC_Ctline00  
                                                                                        
                rdlong  pixels, tile                                                    
                shr     pixels, #1              ' based on LSB, get all desired bits %%1
                xor     pixels, loqbit                                                  
                                                                                        
                and     pixels, loqbit          ' remove unwanted interleaved char      
                shl     pixels, #1              ' alt color %%2                         
                add     screenptr, #2           ' point to next tile                    
                                                                                        
                wrlong  pixels, pixbufptr                                               
                add     pixbufptr, #4                                                   
                djnz    x, #:tile               ' another tile in line?                 
                                                                                        
                jmp     #:loop                                                          

' initialised data and/or presets

opposite        long    %1111 << 18             ' C vs !C

_screen         long    0                       ' @word read-only    
_helpsync       long    0                       ' @word line currently being displayed
_pixbuf         long    0                       ' @2-line long pixel buffer
_user_chars     long    0                       ' base of user defined character set
                                                                     
thisline        long    0                                            
lastline        long    0                                            
                                                                     
lines           long    576                     ' # of pixel lines on screen
loqbit          long    %%1111_1111_1111_1111                        

' uninitialised data and/or temporaries

screenptr       res     1
                         
tileline        res     1
x               res     1
pixbufptr       res     1
tile            res     1
pixels          res     1
temp            res     1

                fit

CON
  columns = 50
  
DAT                     